from typing import (
    Any,
    Dict,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Tuple,
    Union,
    overload,
)

import jax
import jax.numpy as jnp
import numpy as np
from jax.tree_util import register_pytree_node

from ciclo.timetracking import Elapsed
from ciclo.types import CluMetric, LogPath, LogsLike


class Logs(LogsLike):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # copy mutable values
        for k, v in self.items():
            if isinstance(v, MutableMapping):
                self[k] = dict(v)

    @property
    def updates(self) -> Optional[LogsLike]:
        # raise error if accessed
        raise AttributeError("updates is a write-only attribute")

    @updates.setter
    def updates(self, updates: Optional[LogsLike]) -> None:
        if updates is not None:
            self.merge(updates)

    # ----------------------------------
    # logger behavior
    # ----------------------------------
    def add_entry(self, collection: str, name: str, value: Any) -> "Logs":
        if collection not in self:
            self[collection] = {}

        mapping = self[collection]

        if not isinstance(mapping, MutableMapping):
            raise ValueError(
                f"Invalid collection '{collection}' of type '{type(mapping).__name__}', must be a MutableMapping"
            )

        mapping[name] = value

        return self

    def add_entries(self, collection: str, values: Dict[str, Any]) -> "Logs":
        for name, value in values.items():
            self.add_entry(collection, name, value)

        return self

    def add_metric(self, name: str, value: Any) -> "Logs":
        if isinstance(value, CluMetric):
            raise ValueError(
                f"Metric '{name}' is a clu Metric which is stateful, use 'add_stateful_metric' instead"
            )
        return self.add_entry("metrics", name, value)

    def add_metrics(self, metrics: Dict[str, Any]):
        for name, value in metrics.items():
            self.add_metric(name, value)

    def add_stateful_metric(self, name: str, value: Any) -> "Logs":
        return self.add_entry("stateful_metrics", name, value)

    def add_stateful_metrics(self, metrics: Dict[str, Any]) -> "Logs":
        for name, value in metrics.items():
            self.add_stateful_metric(name, value)
        return self

    def add_loss(self, name: str, value: Any, *, add_metric: bool = False) -> "Logs":
        self.add_entry("losses", name, value)
        if add_metric:
            self.add_metric(name, value)
        return self

    def add_losses(
        self, losses: Dict[str, Any], *, add_metrics: bool = False
    ) -> "Logs":
        for name, value in losses.items():
            self.add_loss(name, value, add_metric=add_metrics)
        return self

    def add_output(self, name: str, value: Any, *, per_sample: bool = True) -> "Logs":
        collection = "per_sample_outputs" if per_sample else "outputs"
        self.add_entry(collection, name, value)

    # ----------------------------------
    # history behavior
    # ----------------------------------

    def entry_value(self, name: str) -> Any:
        path = self.entry_path(name)
        if path is None:
            raise KeyError(f"Key {name} not found in logs.")
        collection, name = path
        return self[collection][name]

    def entry_path(self, name: str) -> Optional[LogPath]:
        path = name.split(".")

        if len(path) == 1:
            name = path[0]
            collection = self.entry_collection(name)
            if collection is None:
                return None
        elif len(path) == 2:
            collection, name = path
        else:
            raise ValueError(f"Got more than 2 levels of nesting in key '{name}'")

        return collection, name

    def entry_collection(self, name: str) -> Optional[str]:
        collections = [col for col in self if name in self[col]]

        if len(collections) == 0:
            return None
        elif len(collections) == 1:
            return collections[0]
        else:
            raise ValueError(
                f"Found multiple collections for name '{name}' : {collections}. "
                "Use `collection.name` syntax."
            )

    def merge(self, collection_updates: LogsLike):
        for collection, updates in collection_updates.items():
            if not isinstance(updates, Mapping):
                raise ValueError(
                    f"Invalide value '{updates}' for collection '{collection}', value must be a Mapping"
                )
            if collection in self:
                entries = self[collection]
                if isinstance(entries, MutableMapping):
                    entries.update(updates)
                elif isinstance(entries, Mapping):
                    if type(entries) != type(updates):
                        raise ValueError(
                            f"Cannot merge collections of different types: {type(entries)} and {type(updates)}"
                        )
                    self[collection] = updates
                else:
                    raise ValueError(
                        f"Invalid collection '{collection}' of type '{type(entries).__name__}', must be a Mapping "
                        "or MutableMapping"
                    )
            else:
                # NOTE: we copy mutable mappings to avoid side effects
                if isinstance(updates, Dict):
                    self[collection] = updates.copy()
                elif isinstance(updates, MutableMapping):
                    self[collection] = dict(updates)
                else:
                    self[collection] = updates


def _logs_tree_flatten(self):
    return (dict(self),), None


def _logs_tree_unflatten(aux_data, children):
    self = Logs(children[0])
    return self


register_pytree_node(Logs, _logs_tree_flatten, _logs_tree_unflatten)

# ----------------------------------
# history
# ----------------------------------


class History(List[Logs]):
    @overload
    def collect(self, key: str) -> List[Any]:
        ...

    @overload
    def collect(self, key: str, *keys: str) -> Tuple[List[Any], ...]:
        ...

    def collect(
        self, key: str, *keys: str
    ) -> Union[Logs, List[Any], Tuple[List[Any], ...]]:
        keys = (key,) + keys
        outputs = tuple([] for _ in keys)
        for logs in self:
            paths = [logs.entry_path(key) for key in keys]
            if all(path is not None for path in paths):
                for i, path in enumerate(paths):
                    assert path is not None
                    collection, key = path
                    outputs[i].append(logs[collection][key])

        return outputs if len(keys) > 1 else outputs[0]

    def commit(self, elapsed: Elapsed, logs: LogsLike):
        # convert JAX arrays to numpy arrays to free memory
        logs = jax.tree_map(
            lambda x: np.asarray(x) if isinstance(x, jnp.ndarray) else x, logs
        )
        if not isinstance(logs, Logs):
            logs = Logs(logs)
        logs["elapsed"] = elapsed
        self.append(logs)


__all__ = ["Logs", "History"]



import inspect
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, TypeVar

import jax

from ciclo.logging import Elapsed, History, Logs
from ciclo.loops import LoopFunctionCallback
from ciclo.timetracking import Period
from ciclo.types import A, B, Batch, LogsLike

F = TypeVar("F", bound=Callable[..., Any])


def logs(*args, **kwargs: Dict[str, Any]) -> Logs:
    return Logs(*args, **kwargs)


def history(logs_list: Optional[List[LogsLike]] = None) -> History:
    if logs_list is None:
        return History()

    return History(
        Logs(logs) if not isinstance(logs, Logs) else logs for logs in logs_list
    )


def at(
    steps: Optional[int] = None,
    *,
    samples: Optional[int] = None,
    time: Optional[float] = None,
    date: Optional[float] = None,
) -> Period:
    return Period.create(steps=steps, samples=samples, time=time, date=date)


def is_scalar(x):
    if isinstance(x, (int, float, bool)):
        return True
    elif hasattr(x, "shape"):
        return x.shape == () or x.shape == (1,)
    else:
        return False


def callback(f) -> LoopFunctionCallback:
    return LoopFunctionCallback(f)


def get_batch_size(batch: Batch) -> int:
    def get_size(sizes, x):
        sizes.add(x.shape[0])
        return sizes

    sizes = jax.tree_util.tree_reduce(get_size, batch, set())
    if len(sizes) != 1:
        raise ValueError("Batch size must be the same for all elements in the batch.")
    return sizes.pop()


def elapse(
    dataset: Iterable[B], initial: Optional[Elapsed] = None
) -> Iterable[Tuple[Elapsed, B]]:
    elapsed = initial or Elapsed.create()
    for batch in dataset:
        batch_size = get_batch_size(batch)
        elapsed = elapsed.update(batch_size)
        yield elapsed, batch


def inject(f: Callable[..., A]) -> Callable[..., A]:
    def _inject(*args) -> A:
        n_args = len(inspect.getfullargspec(f).args)
        if inspect.ismethod(f) or inspect.ismethod(f.__call__):
            n_args -= 1
        return f(*args[:n_args])

    return _inject



__version__ = "0.1.0"


from ciclo.types import LogsLike

from . import managed
from .callbacks import (
    checkpoint,
    early_stopping,
    inner_loop,
    keras_bar,
    noop,
    tqdm_bar,
    wandb_logger,
)
from .logging import History, Logs
from .loops import LoopCallbackBase, LoopElement, LoopState, loop
from .schedules import every, piecewise
from .strategies import (
    JIT,
    DataParallel,
    Eager,
    Strategy,
    get_strategy,
    register_strategy,
)
from .timetracking import Elapsed, Period
from .utils import at, callback, elapse, history, inject, logs



from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import Dict, List, Tuple, Union

from ciclo.timetracking import Elapsed, Period
from ciclo.types import Schedule

# ---------------------------------------
# every
# ---------------------------------------


def every(
    steps: Union[int, None] = None,
    *,
    samples: Union[int, None] = None,
    time: Union[timedelta, float, int, None] = None,
    steps_offset: int = 0,
) -> Schedule:
    return Every(
        period=Period.create(steps=steps, samples=samples, time=time),
        last_samples=0,
        last_time=datetime.now().timestamp(),
        steps_offset=steps_offset,
    )


@dataclass(unsafe_hash=True)
class Every:
    period: Period
    last_samples: int
    last_time: float
    steps_offset: int

    def __call__(self, elapsed: Elapsed) -> bool:
        if self.period.steps is not None:
            steps = elapsed.steps - self.steps_offset
            return steps >= 0 and steps % self.period.steps == 0

        if self.period.samples is not None:
            if elapsed.samples - self.last_samples >= self.period.samples:
                self.last_samples = elapsed.samples
                return True

        if self.period.time is not None:
            if elapsed.date - self.last_time >= self.period.time:
                self.last_time = elapsed.date
                return True

        return False


# ---------------------------------------
# piecewise
# ---------------------------------------


def piecewise(
    schedule: Schedule,
    period_schedules: Dict[Period, Schedule],
) -> Schedule:
    return Piecewise(
        schedule=schedule,
        period_schedules=list(period_schedules.items()),
    )


@dataclass
class Piecewise:
    schedule: Schedule
    period_schedules: List[Tuple[Period, Schedule]]

    def __call__(self, elapsed: Elapsed) -> bool:
        if len(self.period_schedules) > 0:
            period, next_schedule = self.period_schedules[0]
            if elapsed >= period:
                self.schedule = next_schedule
                self.period_schedules = self.period_schedules[1:]

        return self.schedule(elapsed)



import functools
import importlib.util
import os
from dataclasses import dataclass, replace
from datetime import datetime
from enum import Enum, auto
from typing import Any, Callable, Dict, Optional, Tuple, Union, overload

from pkbar import Kbar
from tqdm import tqdm

from ciclo.logging import Logs
from ciclo.loops import (
    CallbackOutput,
    LoopCallbackBase,
    LoopOutput,
    LoopState,
    register_adapter,
)
from ciclo.schedules import every
from ciclo.timetracking import Elapsed, Period
from ciclo.types import Batch, LogsLike, S
from ciclo.utils import get_batch_size, is_scalar


def unavailable_dependency(msg: str) -> Any:
    class DependencyNotAvailable(LoopCallbackBase[S]):
        def __init__(self, *args: Any, **kwargs: Any) -> None:
            raise RuntimeError(msg)

        def __loop_callback__(self, loop_state: LoopState[S]) -> CallbackOutput[S]:
            raise RuntimeError(msg)

    return DependencyNotAvailable


class OptimizationMode(str, Enum):
    min = auto()
    max = auto()


class inner_loop(LoopCallbackBase[S]):
    @overload
    def __init__(
        self,
        name_or_loop_fn: str,
        maybe_loop_fn: Callable[[S], LoopOutput[S]],
        *,
        output_state: bool = False,
    ):
        ...

    @overload
    def __init__(
        self,
        name_or_loop_fn: Callable[[S], LoopOutput[S]],
        *,
        output_state: bool = False,
    ):
        ...

    def __init__(
        self,
        name_or_loop_fn: Union[str, Callable[[S], LoopOutput[S]]],
        maybe_loop_fn: Optional[Callable[[S], LoopOutput[S]]] = None,
        *,
        output_state: bool = False,
    ):
        if isinstance(name_or_loop_fn, str):
            assert maybe_loop_fn is not None
            self.name = name_or_loop_fn
            self.loop_fn = maybe_loop_fn
        else:
            assert maybe_loop_fn is None
            self.name = None
            self.loop_fn = name_or_loop_fn
        self.output_state = output_state

    def __call__(self, state: S) -> Tuple[LogsLike, S]:
        inner_state, log_history, _ = self.loop_fn(state)
        logs: LogsLike = log_history[-1] if len(log_history) > 0 else {}
        logs = {
            collection: {
                k + f"_{self.name}" if self.name else k: v
                for k, v in values.items()
                if not isinstance(v, Elapsed)
            }
            for collection, values in logs.items()
            if collection != "elapsed"
        }
        return logs, (inner_state if self.output_state else state)

    def __loop_callback__(self, loop_state: LoopState[S]) -> CallbackOutput[S]:
        return self(loop_state.state)


if importlib.util.find_spec("tensorflow") is not None:
    from flax.training import checkpoints as flax_checkpoints

    class checkpoint(LoopCallbackBase[S]):
        def __init__(
            self,
            ckpt_dir: Union[str, os.PathLike],
            prefix: str = "checkpoint_",
            keep: int = 1,
            overwrite: bool = False,
            keep_every_n_steps: Optional[int] = None,
            async_manager: Optional[flax_checkpoints.AsyncManager] = None,
            monitor: Optional[str] = None,
            mode: Union[str, OptimizationMode] = "min",
        ):
            if isinstance(mode, str):
                mode = OptimizationMode[mode]

            if mode not in OptimizationMode:
                raise ValueError(
                    f"Invalid mode: {mode}, expected one of {list(OptimizationMode)}"
                )
            else:
                self.mode = mode

            self.ckpt_dir = ckpt_dir
            self.prefix = prefix
            self.keep = keep
            self.overwrite = overwrite
            self.keep_every_n_steps = keep_every_n_steps
            self.async_manager = async_manager
            self.monitor = monitor
            self.minimize = self.mode == OptimizationMode.min
            self._best: Optional[float] = None

        def __call__(
            self, elapsed: Elapsed, state: S, logs: Optional[LogsLike] = None
        ) -> None:
            save_checkpoint = True
            step_or_metric = elapsed.steps
            overwrite = self.overwrite

            if self.monitor is not None:
                if logs is None:
                    raise ValueError(
                        "checkpoint callback requires logs to monitor a metric"
                    )
                if not isinstance(logs, Logs):
                    logs = Logs(logs)

                try:
                    value = logs.entry_value(self.monitor)
                except KeyError:
                    raise ValueError(
                        f"Monitored value '{self.monitor}' not found in logs"
                    )

                if (
                    self._best is None
                    or (self.minimize and value < self._best)
                    or (not self.minimize and value > self._best)
                ):
                    self._best = value
                    step_or_metric = (
                        value if self.mode == OptimizationMode.max else -value
                    )
                else:
                    save_checkpoint = False

            if save_checkpoint:
                flax_checkpoints.save_checkpoint(
                    ckpt_dir=self.ckpt_dir,
                    target=state,
                    step=step_or_metric,
                    prefix=self.prefix,
                    keep=self.keep,
                    overwrite=overwrite,
                    keep_every_n_steps=self.keep_every_n_steps,
                    async_manager=self.async_manager,
                )

        def __loop_callback__(self, loop_state: LoopState[S]) -> CallbackOutput[S]:
            self(loop_state.elapsed, loop_state.state, loop_state.accumulated_logs)
            return {}, loop_state.state

else:
    checkpoint = unavailable_dependency(
        "'tensorflow' package is not available, please install it to use the 'checkpoint' callback"
    )


class early_stopping(LoopCallbackBase[S]):
    def __init__(
        self,
        monitor: str,
        patience: Union[int, Period],
        min_delta: float = 0,
        mode: Union[str, OptimizationMode] = "min",
        baseline: Optional[float] = None,
        restore_best_weights: bool = False,
    ):
        if isinstance(mode, str):
            mode = OptimizationMode[mode]

        if mode not in OptimizationMode:
            raise ValueError(
                f"Invalid mode: {mode}, expected one of {list(OptimizationMode)}"
            )
        else:
            self.mode = mode

        self.monitor = monitor
        self.patience = (
            patience if isinstance(patience, Period) else Period.create(patience)
        )
        self.min_delta = min_delta
        self.mode = mode
        self.baseline = baseline
        self.restore_best_weights = restore_best_weights
        self.minimize = self.mode == OptimizationMode.min
        self._best = baseline
        self._best_state = None
        self._elapsed_start: Optional[Elapsed] = None

    def __call__(self, elapsed: Elapsed, state: S, logs: LogsLike) -> Tuple[bool, S]:
        stop_iteration = False

        if self._elapsed_start is None:
            self._elapsed_start = elapsed

        if not isinstance(logs, Logs):
            logs = Logs(logs)

        try:
            value = logs.entry_value(self.monitor)
        except KeyError:
            raise ValueError(f"Monitored value '{self.monitor}' not found in logs")

        if (
            self._best is None
            or (self.minimize and value < self._best)
            or (not self.minimize and value > self._best)
        ):
            self._best = value
            self._best_state = state
            self._elapsed_start = elapsed

        if elapsed - self._elapsed_start >= self.patience:
            if self.restore_best_weights and self._best_state is not None:
                state = self._best_state
            stop_iteration = True

        return stop_iteration, state

    def __loop_callback__(self, loop_state: LoopState[S]) -> CallbackOutput[S]:
        stop_iteration, state = self(
            loop_state.elapsed, loop_state.state, loop_state.accumulated_logs
        )
        loop_state.stop_iteration = stop_iteration
        return {}, state


class tqdm_bar(LoopCallbackBase[S]):
    def __init__(
        self,
        total: Union[Period, int, None] = None,
        desc=None,
        leave=True,
        file=None,
        ncols=None,
        mininterval=0.1,
        maxinterval=10.0,
        miniters=None,
        ascii=None,
        disable=False,
        unit_scale=False,
        dynamic_ncols=False,
        smoothing=0.3,
        bar_format=None,
        initial=0,
        position=None,
        postfix=None,
        unit_divisor=1000,
        write_bytes=None,
        lock_args=None,
        nrows=None,
        colour=None,
        delay=0,
        gui=False,
        **kwargs,
    ):
        if isinstance(total, int):
            total = Period.create(steps=total)

        if total is not None:
            if total.steps is not None:
                bar_total = total.steps
                unit = "steps"
            elif total.samples is not None:
                bar_total = total.samples
                unit = "samples"
            elif total.time is not None:
                bar_total = total.time
                unit = "s"
                unit_scale = True
            elif total.date is not None:
                total = replace(total, time=total.date - datetime.now().timestamp())
                bar_total = total.time
                unit = "s"
                unit_scale = True
            else:
                raise ValueError("Invalid total")
        else:
            bar_total = None
            unit = "it"

        self.total = total
        self.prev_step: Optional[int] = None
        self.prev_samples: Optional[int] = None
        self.prev_time: Optional[float] = None
        self.bar_total = bar_total
        self.bar = tqdm(
            desc=desc,
            total=bar_total,
            leave=leave,
            file=file,
            ncols=ncols,
            mininterval=mininterval,
            maxinterval=maxinterval,
            miniters=miniters,
            ascii=ascii,
            disable=disable,
            unit=unit,
            unit_scale=unit_scale,
            dynamic_ncols=dynamic_ncols,
            smoothing=smoothing,
            bar_format=bar_format,
            initial=initial,
            position=position,
            postfix=postfix,
            unit_divisor=unit_divisor,
            write_bytes=write_bytes,
            lock_args=lock_args,
            nrows=nrows,
            colour=colour,
            delay=delay,
            gui=gui,
            **kwargs,
        )

    def __call__(self, elapsed: Elapsed, batch: Optional[Batch] = None) -> None:
        if self.total is None or self.total.steps is not None:
            if self.prev_step is None:
                self.prev_step = elapsed.steps - 1
            self.bar.update(elapsed.steps - self.prev_step)
            self.prev_step = elapsed.steps
        elif self.total.samples is not None:
            if self.prev_samples is None:
                if batch is None:
                    raise ValueError("Batch is required for sample-based progress bar")
                self.prev_samples = elapsed.samples - get_batch_size(batch)
            self.bar.update(elapsed.samples - self.prev_samples)
            self.prev_samples = elapsed.samples
        elif self.total.time is not None:
            if self.prev_time is None:
                self.prev_time = elapsed._date_start
            self.bar.update(elapsed.date - self.prev_time)
            self.prev_time = elapsed.date
        else:
            raise ValueError("Invalid total")

    def __loop_callback__(self, loop_state: LoopState[S]) -> CallbackOutput[S]:
        self(loop_state.elapsed, loop_state.batch)
        return {}, loop_state.state


class keras_bar(LoopCallbackBase[S]):
    def __init__(
        self,
        total: Union[Period, int, None] = None,
        epoch=None,
        num_epochs=None,
        width=30,
        verbose=1,
        interval=0.05,
        stateful_metrics=None,
        always_stateful=False,
        unit_name="step",
    ):
        if isinstance(total, int):
            total = Period.create(steps=total)

        if total is not None:
            if total.steps is not None:
                bar_total = total.steps
                unit_name = "step"
            elif total.samples is not None:
                bar_total = total.samples
                unit_name = "sample"
            elif total.time is not None:
                bar_total = total.time
                unit_name = "s"
            elif total.date is not None:
                total = replace(total, time=total.date - datetime.now().timestamp())
                bar_total = total.time
                unit_name = "s"
            else:
                raise ValueError("Invalid total")
        else:
            bar_total = None
            unit_name = "it"

        self.total = total
        self.prev_step: Optional[int] = None
        self.prev_samples: Optional[int] = None
        self.prev_time: Optional[float] = None
        self.bar_total = bar_total
        self.bar = Kbar(
            bar_total,
            epoch=epoch,
            num_epochs=num_epochs,
            width=width,
            verbose=verbose,
            interval=interval,
            stateful_metrics=stateful_metrics,
            always_stateful=always_stateful,
            unit_name=unit_name,
        )

    def __call__(self, elapsed: Elapsed, logs: LogsLike) -> None:
        if self.total is None or self.total.steps is not None:
            current = elapsed.steps
        elif self.total.samples is not None:
            current = elapsed.samples
        elif self.total.time is not None:
            current = elapsed.time
        else:
            raise ValueError("Invalid total")

        metrics: Dict[str, Any] = {}
        if "stateful_metrics" in logs:
            stateful_metrics = logs["stateful_metrics"]
            self.bar.stateful_metrics.update(stateful_metrics.keys())
            metrics.update(stateful_metrics)

        if "metrics" in logs:
            metrics.update(logs["metrics"])

        if metrics:
            self.bar.update(
                current,
                values=[(k, v) for k, v in metrics.items() if is_scalar(v)],
            )

    def __loop_callback__(self, loop_state: LoopState[S]) -> CallbackOutput[S]:
        self(loop_state.elapsed, loop_state.logs)
        return {}, loop_state.state


if importlib.util.find_spec("wandb") is not None:
    from wandb.wandb_run import Run

    class wandb_logger(LoopCallbackBase[S]):
        def __init__(self, run: Run):
            from wandb.wandb_run import Run

            self.run: Run = run

        def __call__(self, elapsed: Elapsed, logs: LogsLike) -> None:
            data = {}
            for collection, collection_logs in logs.items():
                for key, value in collection_logs.items():
                    if is_scalar(value):
                        if key in data:
                            key = f"{collection}.{key}"
                        data[key] = value

            if len(data) > 0:
                self.run.log(data, step=elapsed.steps)

        def __loop_callback__(self, loop_state: LoopState[S]) -> CallbackOutput[S]:
            self(loop_state.elapsed, loop_state.logs)
            return {}, loop_state.state

else:
    wandb_logger = unavailable_dependency(
        "'wandb' package is not available, please install it to use the 'wandb_logger' callback"
    )


class NoOp(LoopCallbackBase[S]):
    def __loop_callback__(self, loop_state: LoopState[S]) -> CallbackOutput[S]:
        return {}, loop_state.state


noop = NoOp()

# -------------------------------------------
# Adapters
# -------------------------------------------

if importlib.util.find_spec("clu") is not None:
    from clu.periodic_actions import PeriodicAction

    @dataclass(frozen=True)
    class PeriodicActionCallback(LoopCallbackBase[S]):
        action: PeriodicAction

        def __loop_callback__(self, loop_state: LoopState[S]) -> CallbackOutput[S]:
            self.action(loop_state.elapsed.steps, t=loop_state.elapsed.date)
            return {}, loop_state.state

    @functools.partial(register_adapter, cls=PeriodicAction)
    def periodic_action_adapter(f: PeriodicAction):
        return PeriodicActionCallback(f)



from abc import ABC, abstractmethod
from dataclasses import dataclass, replace
from typing import Any, Callable, Dict, Optional, TypeVar, overload

import jax
import jax.numpy as jnp
from einop import einop
from flax import jax_utils
from typing_extensions import Protocol, runtime_checkable

from ciclo.loops import GeneralCallback
from ciclo.types import CluMetric


class Dataclass(Protocol):
    __dataclass_fields__: Dict


@runtime_checkable
class HasKey(Protocol):
    key: jax.random.KeyArray


StrategyConstructor = Callable[[], "Strategy"]
A = TypeVar("A")
S = TypeVar("S", bound=Dataclass)
ME = TypeVar("ME", bound=CluMetric)


_REGISTRY: Dict[str, StrategyConstructor] = {}


# ----------------------------------------------------------------------------
# utils
# ----------------------------------------------------------------------------


@overload
def register_strategy(
    name: str,
) -> Callable[[StrategyConstructor], StrategyConstructor]:
    ...


@overload
def register_strategy(
    name: str,
    *,
    constructor: StrategyConstructor,
) -> None:
    ...


def register_strategy(
    name: str,
    *,
    constructor: Optional[StrategyConstructor] = None,
) -> Optional[Callable[[StrategyConstructor], StrategyConstructor]]:
    """
    Register a strategy class.
    """

    def _register(constructor: StrategyConstructor):
        if name in _REGISTRY:
            raise ValueError(f"Strategy {name} already registered")

        _REGISTRY[name] = constructor

    if constructor is None:

        def decorator(
            constructor: StrategyConstructor,
        ) -> StrategyConstructor:
            _register(constructor)
            return constructor

        return decorator
    else:
        _register(constructor)


def get_strategy(name: str) -> "Strategy":
    """
    Get a strategy class.
    """
    if name not in _REGISTRY:
        raise ValueError(f"Strategy {name} not registered")

    return _REGISTRY[name]()


# ----------------------------------------------------------------------------
# register strategies
# ----------------------------------------------------------------------------

register_strategy(
    name="eager",
    constructor=lambda: Eager(),
)
register_strategy(
    name="jit",
    constructor=lambda: JIT(donate_args=False),
)
register_strategy(
    name="jit_donate",
    constructor=lambda: JIT(donate_args=True),
)
register_strategy(
    name="data_parallel",
    constructor=lambda: DataParallel(donate_args=False),
)
register_strategy(
    name="data_parallel_donate",
    constructor=lambda: DataParallel(donate_args=True),
)


# ----------------------------------------------------------------------------
# Strategy
# ----------------------------------------------------------------------------


class Strategy(ABC):
    def from_host(self, state: S) -> S:
        return state

    def to_host(self, state: S) -> S:
        return state

    def lift_batch(self, data: A) -> A:
        return data

    def lift_key(self, key: jax.random.KeyArray) -> jax.random.KeyArray:
        return key

    def lift_batch_size(self, batch_size: int) -> int:
        return batch_size

    def handle_metric(self, metric: ME) -> ME:
        return metric

    def handle_grads(
        self,
        grads: Any,
    ) -> Any:
        return grads

    def handle_batch_stats(
        self,
        batch_stats: Any,
    ) -> Any:
        return batch_stats

    def lower_tileable(self, logs: A) -> A:
        return jax.device_get(logs)

    def lower_sharded(self, logs: A) -> A:
        return jax.device_get(logs)

    def lower_averageable(self, logs: A) -> A:
        return jax.device_get(logs)

    def lower_replicated(self, logs: A) -> A:
        return jax.device_get(logs)

    @abstractmethod
    def __call__(self, callback: GeneralCallback[S]) -> GeneralCallback[S]:
        ...


@dataclass(eq=True, frozen=True)
class Eager(Strategy):
    def __call__(self, callback: GeneralCallback[S]) -> GeneralCallback[S]:
        return callback


@dataclass(eq=True, frozen=True)
class JIT(Strategy):
    donate_args: bool = False

    def __call__(self, callback: GeneralCallback[S]) -> GeneralCallback[S]:
        return jax.jit(
            callback,
            donate_argnums=0 if self.donate_args else (),
            static_argnums=3,
        )


@dataclass(eq=True, frozen=True)
class DataParallel(Strategy):
    axis_name: str = "device"
    donate_args: bool = False

    def from_host(self, state: S) -> S:
        key = state.key if isinstance(state, HasKey) else None
        state = jax_utils.replicate(state)
        if key is not None:
            devices = jax_utils._pmap_device_order()
            key = jax.random.split(key, jax.local_device_count())
            key = jax.device_put_sharded(list(key), devices)
            state = replace(state, key=key)
        return state

    def to_host(self, state: S) -> S:
        return jax_utils.unreplicate(state)

    def lift_batch(self, data: A) -> A:
        data = jax.tree_map(
            lambda x: einop(
                x,
                "(device batch) ... -> device batch ...",
                device=jax.local_device_count(),
            ),
            data,
        )
        return data

    def lift_key(self, key: jax.random.KeyArray) -> jax.random.KeyArray:
        return jax.random.split(key, jax.local_device_count())

    def lift_batch_size(self, batch_size: int) -> int:
        return batch_size * jax.local_device_count()

    def handle_metric(
        self,
        metric: ME,
    ) -> ME:
        # metrics = jax.lax.stop_gradient(metrics)
        metric = jax.lax.all_gather(metric, axis_name=self.axis_name)
        metric = metric.reduce()
        return metric

    def handle_grads(self, grads: A) -> A:
        return jax.lax.pmean(grads, axis_name=self.axis_name)

    def handle_batch_stats(self, batch_stats: A) -> A:
        return jax.lax.pmean(batch_stats, axis_name=self.axis_name)

    def lower_averageable(self, logs: A) -> A:
        return jax.tree_util.tree_map(lambda x: jnp.mean(x, axis=0), logs)

    def lower_tileable(self, logs: A) -> A:
        return jax.tree_util.tree_map(
            lambda x: einop(x, "device batch ... -> (device batch) ..."), logs
        )

    def lower_sharded(self, logs: A) -> A:
        return jax.device_get(logs)

    def lower_replicated(self, logs: A) -> A:
        return jax.tree_util.tree_map(lambda x: x[0], logs)

    def __call__(self, callback: GeneralCallback[S]) -> GeneralCallback[S]:
        return jax.pmap(
            callback,
            axis_name=self.axis_name,
            donate_argnums=0 if self.donate_args else (),
            in_axes=(0, 0, None, None),
            out_axes=(0, 0),
            static_broadcasted_argnums=3,
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ciclo']

package_data = \
{'': ['*']}

install_requires = \
['einop>=0.0.1',
 'flax>=0.6.0',
 'jax>=0.3.0',
 'jaxlib>=0.3.0',
 'pkbar>=0.5',
 'tqdm>=4.0.0']

setup_kwargs = {
    'name': 'ciclo',
    'version': '0.1.5',
    'description': '',
    'long_description': 'None',
    'author': 'Cristian Garcia',
    'author_email': 'cgarcia.e88@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



import inspect
from abc import abstractmethod
from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    List,
    Optional,
    Tuple,
    Type,
    Union,
)

from typing_extensions import Protocol, runtime_checkable

import ciclo
from ciclo.logging import History, Logs
from ciclo.timetracking import Elapsed, Period
from ciclo.types import (
    A,
    B,
    Batch,
    Broadcasts,
    InputCallback,
    LogsLike,
    S,
    Schedule,
    Statics,
)

CallbackOutput = Tuple[LogsLike, S]


@runtime_checkable
class LoopCallback(Protocol, Generic[S]):
    def __loop_callback__(self, loop_state: "LoopState[S]") -> CallbackOutput[S]:
        ...


InputTasks = Dict[Schedule, Union[InputCallback, List[InputCallback]]]
ScheduleCallback = Dict[Schedule, List[LoopCallback[S]]]
CallbackAdapter = Callable[[Any], LoopCallback[S]]

FunctionCallbackOutputs = Union[
    Tuple[Optional[LogsLike], Optional[S]], LogsLike, S, None
]
GeneralCallback = Callable[[S, Batch, Broadcasts, Statics], FunctionCallbackOutputs[S]]


@dataclass
class LoopState(Generic[S]):
    state: S
    batch: Batch
    history: History
    elapsed: Elapsed
    logs: Logs
    accumulated_logs: Logs
    metadata: Any
    stop_iteration: bool = False


LoopOutput = Tuple[S, History, Elapsed]


def to_standard_outputs(
    outputs: FunctionCallbackOutputs[S], current_state: S
) -> CallbackOutput[S]:
    logs: LogsLike
    state: S
    if outputs is None:
        logs = {}
        state = current_state

    elif type(outputs) is tuple:
        if len(outputs) != 2:
            raise ValueError(
                f"Invalid output from callback function: {outputs}, must be a tuple of length 2"
            )
        logs = outputs[0] or {}
        state = outputs[1] or current_state
    elif isinstance(outputs, Dict):
        logs = outputs
        state = current_state
    else:
        logs = {}
        state = outputs

    return logs, state


class LoopElement:
    def keys(self) -> List[Schedule]:
        return [ciclo.every(1)]

    def __getitem__(self: A, key: Schedule) -> List[A]:
        return [self]


class LoopCallbackBase(LoopCallback[S], LoopElement):
    @abstractmethod
    def __loop_callback__(self, loop_state: LoopState[S]) -> CallbackOutput[S]:
        ...


@dataclass(frozen=True)
class LoopFunctionCallback(LoopCallbackBase[S]):
    f: Callable[..., FunctionCallbackOutputs[S]]

    def __loop_callback__(self, loop_state: LoopState[S]) -> CallbackOutput[S]:
        outputs = ciclo.inject(self.f)(
            loop_state.state, loop_state.batch, loop_state.elapsed, loop_state
        )
        return to_standard_outputs(outputs, loop_state.state)


def loop(
    state: S,
    dataset: Iterable[B],
    tasks: InputTasks,
    *,
    stop: Union[Period, int, None] = None,
    on_start: Union[InputCallback, List[InputCallback], None] = None,
    on_end: Union[InputCallback, List[InputCallback], None] = None,
    history: Optional[History] = None,
    elapsed: Optional[Elapsed] = None,
    catch_keyboard_interrupt: bool = True,
    metadata: Optional[Any] = None,
) -> LoopOutput[S]:
    if isinstance(stop, int):
        stop_period = Period.create(steps=stop)
    else:
        stop_period = stop

    if history is None:
        history = History()

    loop_state = LoopState(
        state=state,
        batch=None,
        history=history,
        elapsed=elapsed or Elapsed.create(),
        logs=Logs(),
        accumulated_logs=Logs(),
        metadata=metadata,
    )

    if on_start is None:
        on_start = []
    elif not isinstance(on_start, list):
        on_start = [on_start]

    if on_end is None:
        on_end = []
    elif not isinstance(on_end, list):
        on_end = [on_end]

    tasks_ = [
        (
            schedule,
            [
                get_loop_callback(f)
                for f in (callbacks if isinstance(callbacks, list) else [callbacks])
            ],
        )
        for schedule, callbacks in tasks.items()
    ]

    try:
        for i, (elapsed, batch) in enumerate(ciclo.elapse(dataset, initial=elapsed)):
            loop_state.elapsed = elapsed
            loop_state.batch = batch

            # call on_start on first batch
            if i == 0:
                loop_state.logs = Logs()
                for callback in on_start:
                    callback = get_loop_callback(callback)
                    _make_call(loop_state, callback)

            loop_state.logs = Logs()
            for s, (schedule, callbacks) in enumerate(tasks_):
                if schedule(loop_state.elapsed):
                    for callback in callbacks:
                        _make_call(loop_state, callback)
                        if loop_state.stop_iteration:
                            break
                if loop_state.stop_iteration:
                    break

            if loop_state.logs:
                loop_state.history.commit(elapsed, loop_state.logs)

            if loop_state.stop_iteration or (
                stop_period and loop_state.elapsed >= stop_period
            ):
                break

        # call on_end on last batch
        loop_state.logs = Logs()
        for callback in on_end:
            callback = get_loop_callback(callback)
            _make_call(loop_state, callback)

    except KeyboardInterrupt:
        if catch_keyboard_interrupt:
            print("\nStopping loop...")
        else:
            raise

    return loop_state.state, loop_state.history, loop_state.elapsed


def _make_call(loop_state: LoopState[S], callback: LoopCallback[S]):
    try:
        loop_state.elapsed = loop_state.elapsed.update_time()
        logs, state = callback.__loop_callback__(loop_state)
        loop_state.logs.merge(logs)
        loop_state.accumulated_logs.merge(logs)
        loop_state.state = state
    except BaseException as e:
        raise type(e)(f"Error in callback {callback}: {e}") from e


# ---------------------------------------
# Registry
# ---------------------------------------
_REGISTRY: Dict[Type, CallbackAdapter] = {}


def register_adapter(adapter: CallbackAdapter, cls: Type):
    if cls in _REGISTRY:
        raise ValueError(f"Adapter for {cls} already registered")

    _REGISTRY[cls] = adapter


def get_adapter(cls: Type) -> Optional[CallbackAdapter]:
    super_classes = inspect.getmro(cls)
    for super_class in super_classes:
        if super_class in _REGISTRY:
            return _REGISTRY[super_class]
    return None


def get_loop_callback(f: Any) -> LoopCallback:
    adaper = get_adapter(type(f))

    if adaper is not None:
        return adaper(f)
    elif isinstance(f, LoopCallbackBase):
        return f
    elif callable(f):
        return LoopFunctionCallback(f)
    else:
        raise ValueError(f"Invalid callback: {f}")



import importlib.util
from typing import Any, Callable, Dict, Mapping, Tuple, TypeVar

import ciclo

if importlib.util.find_spec("clu"):
    from clu.metrics import Metric as CluMetric
else:
    locals()["CluMetric"] = type("CluMetric", (), {})

State = Any
Batch = Any
Broadcasts = Any
Statics = Any
LogPath = Tuple[str, str]
LogsLike = Dict[str, Mapping[str, Any]]
InputCallback = Any
A = TypeVar("A")
S = TypeVar("S", bound=State)
B = TypeVar("B", bound=Batch)
Schedule = Callable[["ciclo.Elapsed"], bool]
